"""A Python toolkit for machine learning focused on regression tasks.

This package provides models, feature selection, visualization, metrics, and other utilities
for machine learning workflows.
"""

__version__ = "0.4.0-rc.0"
__license__ = "Apache-2.0"
__author__ = "yu9824"
__copyright__ = "Copyright © 2025 yu9824"
